"""add-cancelled-to-kernelstatus

Revision ID: 513164749de4
Revises: 405aa2c39458
Create Date: 2019-09-20 11:13:39.157834

"""
import textwrap

import sqlalchemy as sa
from alembic import op
from sqlalchemy.dialects import postgresql
from sqlalchemy.sql import text

# revision identifiers, used by Alembic.
revision = "513164749de4"
down_revision = "405aa2c39458"
branch_labels = None
depends_on = None

kernelstatus_new_values = [
    "PENDING",
    "PREPARING",
    "BUILDING",
    "PULLING",
    "RUNNING",
    "RESTARTING",
    "RESIZING",
    "SUSPENDED",
    "TERMINATING",
    "TERMINATED",
    "ERROR",
    "CANCELLED",  # added
]
kernelstatus_new = postgresql.ENUM(*kernelstatus_new_values, name="kernelstatus")

kernelstatus_old_values = [
    "PENDING",
    "PREPARING",
    "BUILDING",
    "PULLING",
    "RUNNING",
    "RESTARTING",
    "RESIZING",
    "SUSPENDED",
    "TERMINATING",
    "TERMINATED",
    # 'ERROR',
]
kernelstatus_old = postgresql.ENUM(*kernelstatus_old_values, name="kernelstatus")


def upgrade():
    conn = op.get_bind()
    conn.execute(text("DROP INDEX IF EXISTS ix_kernels_unique_sess_token;"))
    conn.execute(text("ALTER TYPE kernelstatus RENAME TO kernelstatus_old;"))
    kernelstatus_new.create(conn)
    query = textwrap.dedent("""
    ALTER TABLE kernels
        ALTER COLUMN "status" DROP DEFAULT,
        ALTER COLUMN "status" TYPE kernelstatus USING "status"::text::kernelstatus,
        ALTER COLUMN "status" SET DEFAULT 'PENDING'::kernelstatus;
    """)
    conn.execute(text(query))
    query = "DROP TYPE kernelstatus_old;"
    conn.execute(text(query))
    op.create_index(
        "ix_kernels_unique_sess_token",
        "kernels",
        ["access_key", "sess_id"],
        unique=True,
        postgresql_where=sa.text("status NOT IN ('TERMINATED', 'CANCELLED') and role = 'master'"),
    )


def downgrade():
    op.drop_index("ix_kernels_unique_sess_token", table_name="kernels")
    conn = op.get_bind()
    conn.execute(text("ALTER TYPE kernelstatus RENAME TO kernelstatus_new;"))
    kernelstatus_old.create(conn)
    conn.execute(text(textwrap.dedent("""\
            ALTER TABLE kernels
                ALTER COLUMN "status" DROP DEFAULT,
                ALTER COLUMN "status" TYPE kernelstatus USING (
                    CASE "status"::text
                        WHEN 'CANCELLED' THEN 'TERMINATED'
                        ELSE "status"::text
                    END
                )::kernelstatus,
                ALTER COLUMN "status" SET DEFAULT 'PREPARING'::kernelstatus;
            """)))
    conn.execute(text("DROP TYPE kernelstatus_new;"))
    op.create_index(
        "ix_kernels_unique_sess_token",
        "kernels",
        ["access_key", "sess_id"],
        unique=True,
        postgresql_where=sa.text("status != 'TERMINATED' and role = 'master'"),
    )



from __future__ import annotations

import logging
import sys
import traceback
from typing import TYPE_CHECKING, Any, Mapping

from ai.backend.common.events import AgentErrorEvent
from ai.backend.common.logging import BraceStyleAdapter
from ai.backend.common.plugin.monitor import AbstractErrorReporterPlugin
from ai.backend.common.types import AgentId, LogSeverity

from ..models import error_logs

if TYPE_CHECKING:
    from ai.backend.manager.api.context import RootContext

log = BraceStyleAdapter(logging.getLogger(__spec__.name))  # type: ignore[name-defined]


class ErrorMonitor(AbstractErrorReporterPlugin):
    async def init(self, context: Any = None) -> None:
        if context is None:
            log.warning(
                (
                    "manager.plugin.error_monitor is initialized without the root context. "
                    "The plugin is disabled."
                ),
            )
            self.enabled = False
            return
        else:
            self.enabled = True
        root_ctx: RootContext = context["_root.context"]  # type: ignore
        self.event_dispatcher = root_ctx.event_dispatcher
        self._evh = self.event_dispatcher.consume(AgentErrorEvent, None, self.handle_agent_error)
        self.db = root_ctx.db

    async def cleanup(self) -> None:
        if self.enabled:
            self.event_dispatcher.unconsume(self._evh)

    async def update_plugin_config(self, plugin_config: Mapping[str, Any]) -> None:
        pass

    async def capture_message(self, message: str) -> None:
        pass

    async def capture_exception(
        self,
        exc_instance: Exception = None,
        context: Mapping[str, Any] = None,
    ) -> None:
        if not self.enabled:
            return
        if exc_instance:
            tb = exc_instance.__traceback__
        else:
            _, sys_exc_instance, tb = sys.exc_info()
            if isinstance(sys_exc_instance, BaseException) and not isinstance(
                sys_exc_instance, Exception
            ):
                # bypass BaseException as they are used for controlling the process/coroutine lifecycles
                # instead of indicating actual errors
                return
            exc_instance = sys_exc_instance
        exc_type: Any = type(exc_instance)

        if context is None or "severity" not in context:
            severity = LogSeverity.ERROR
        else:
            severity = context["severity"]
        if context is None or "user" not in context:
            user = None
        else:
            user = context["user"]
        message = "".join(traceback.format_exception_only(exc_type, exc_instance)).strip()

        async with self.db.begin() as conn:
            query = error_logs.insert().values(
                {
                    "severity": severity,
                    "source": "manager",
                    "user": user,
                    "message": message,
                    "context_lang": "python",
                    "context_env": context,
                    "traceback": "".join(traceback.format_tb(tb)).strip(),
                }
            )
            await conn.execute(query)
        log.debug(
            'collected an error log [{}] "{}" from manager',
            severity.name,
            message,
        )

    async def handle_agent_error(
        self,
        context: None,
        source: AgentId,
        event: AgentErrorEvent,
    ) -> None:
        if not self.enabled:
            return
        async with self.db.begin() as conn:
            query = error_logs.insert().values(
                {
                    "severity": event.severity,
                    "source": source,
                    "user": event.user,
                    "message": event.message,
                    "context_lang": "python",
                    "context_env": event.context_env,
                    "traceback": event.traceback,
                }
            )
            await conn.execute(query)
        log.debug(
            'collected an error log [{}] "{}" from agent:{}',
            event.severity.name,
            event.message,
            source,
        )



from __future__ import annotations

import logging
from typing import TYPE_CHECKING, Any, AsyncGenerator, Iterable, Mapping, Tuple

import aiohttp_cors
import trafaret as t
from aiohttp import web

from ai.backend.common.docker import get_known_registries
from ai.backend.common.logging import BraceStyleAdapter

from .auth import superadmin_required
from .exceptions import InvalidAPIParameters
from .types import CORSOptions, WebMiddleware
from .utils import check_api_params

if TYPE_CHECKING:
    from .context import RootContext

log = BraceStyleAdapter(logging.getLogger(__spec__.name))  # type: ignore[name-defined]


async def get_resource_slots(request: web.Request) -> web.Response:
    log.info("ETCD.GET_RESOURCE_SLOTS ()")
    root_ctx: RootContext = request.app["_root.context"]
    known_slots = await root_ctx.shared_config.get_resource_slots()
    return web.json_response(known_slots, status=200)


async def get_vfolder_types(request: web.Request) -> web.Response:
    log.info("ETCD.GET_VFOLDER_TYPES ()")
    root_ctx: RootContext = request.app["_root.context"]
    vfolder_types = await root_ctx.shared_config.get_vfolder_types()
    return web.json_response(vfolder_types, status=200)


@superadmin_required
async def get_docker_registries(request: web.Request) -> web.Response:
    """
    Returns the list of all registered docker registries.
    """
    log.info("ETCD.GET_DOCKER_REGISTRIES ()")
    root_ctx: RootContext = request.app["_root.context"]
    _registries = await get_known_registries(root_ctx.shared_config.etcd)
    # ``yarl.URL`` is not JSON-serializable, so we need to represent it as string.
    known_registries: Mapping[str, str] = {k: v.human_repr() for k, v in _registries.items()}
    return web.json_response(known_registries, status=200)


@superadmin_required
@check_api_params(
    t.Dict(
        {
            t.Key("key"): t.String,
            t.Key("prefix", default=False): t.Bool,
        }
    )
)
async def get_config(request: web.Request, params: Any) -> web.Response:
    root_ctx: RootContext = request.app["_root.context"]
    log.info(
        "ETCD.GET_CONFIG (ak:{}, key:{}, prefix:{})",
        request["keypair"]["access_key"],
        params["key"],
        params["prefix"],
    )
    if params["prefix"]:
        # Flatten the returned ChainMap object for JSON serialization
        tree_value = dict(await root_ctx.shared_config.etcd.get_prefix_dict(params["key"]))
        return web.json_response({"result": tree_value})
    else:
        scalar_value = await root_ctx.shared_config.etcd.get(params["key"])
        return web.json_response({"result": scalar_value})


@superadmin_required
@check_api_params(
    t.Dict(
        {
            t.Key("key"): t.String,
            t.Key("value"): t.String(allow_blank=True) | t.Mapping(
                t.String(allow_blank=True), t.Any
            ),
        }
    )
)
async def set_config(request: web.Request, params: Any) -> web.Response:
    root_ctx: RootContext = request.app["_root.context"]
    log.info(
        "ETCD.SET_CONFIG (ak:{}, key:{}, val:{})",
        request["keypair"]["access_key"],
        params["key"],
        params["value"],
    )
    if isinstance(params["value"], Mapping):
        updates = {}

        def flatten(prefix, o):
            for k, v in o.items():
                inner_prefix = prefix if k == "" else f"{prefix}/{k}"
                if isinstance(v, Mapping):
                    flatten(inner_prefix, v)
                else:
                    updates[inner_prefix] = v

        flatten(params["key"], params["value"])
        # TODO: chunk support if there are too many keys
        if len(updates) > 16:
            raise InvalidAPIParameters("Too large update! Split into smaller key-value pair sets.")
        await root_ctx.shared_config.etcd.put_dict(updates)
    else:
        await root_ctx.shared_config.etcd.put(params["key"], params["value"])
    return web.json_response({"result": "ok"})


@superadmin_required
@check_api_params(
    t.Dict(
        {
            t.Key("key"): t.String,
            t.Key("prefix", default=False): t.Bool,
        }
    )
)
async def delete_config(request: web.Request, params: Any) -> web.Response:
    root_ctx: RootContext = request.app["_root.context"]
    log.info(
        "ETCD.DELETE_CONFIG (ak:{}, key:{}, prefix:{})",
        request["keypair"]["access_key"],
        params["key"],
        params["prefix"],
    )
    if params["prefix"]:
        await root_ctx.shared_config.etcd.delete_prefix(params["key"])
    else:
        await root_ctx.shared_config.etcd.delete(params["key"])
    return web.json_response({"result": "ok"})


async def app_ctx(app: web.Application) -> AsyncGenerator[None, None]:
    root_ctx: RootContext = app["_root.context"]
    if root_ctx.pidx == 0:
        await root_ctx.shared_config.register_myself()
    yield
    if root_ctx.pidx == 0:
        await root_ctx.shared_config.deregister_myself()


def create_app(
    default_cors_options: CORSOptions,
) -> Tuple[web.Application, Iterable[WebMiddleware]]:
    app = web.Application()
    app.cleanup_ctx.append(app_ctx)
    app["prefix"] = "config"
    app["api_versions"] = (3, 4)
    cors = aiohttp_cors.setup(app, defaults=default_cors_options)
    cors.add(app.router.add_route("GET", r"/resource-slots", get_resource_slots))
    cors.add(app.router.add_route("GET", r"/vfolder-types", get_vfolder_types))
    cors.add(app.router.add_route("GET", r"/docker-registries", get_docker_registries))
    cors.add(app.router.add_route("POST", r"/get", get_config))
    cors.add(app.router.add_route("POST", r"/set", set_config))
    cors.add(app.router.add_route("POST", r"/delete", delete_config))
    return app, []
